"""This reflects a patient data set structure
which consumes multiple datasetinstances in order
to link them by their reference number"""

from .mtbconverter_exceptions import NoReferenceIdFound

class PatientDataSet():

    def __init__(self, qbic_pat_id, qbic_sample_id,
        datetime, datasetinstance_list):
        self._qbic_pat_id = qbic_pat_id
        self._qbic_sample_id = qbic_sample_id
        self._datetime = datetime
        self._datasetinstance_list = datasetinstance_list

    def patientdataset(self):
        # Create the patient data set
        patient_data_set = cx.PatientDataSetType(Source='XMLIMPORT',
            OrganisationUnitRefs='QBIC')
        
        # Create ID container with QBIC patient id first
        id_container = cx.IDContainerType()
        id_flexible = cx.FlexibleIDType(key='QBIC_PAT_ID',
            value=self._qbic_pat_id)
        id_container.append(id_flexible)
        patient_data_set.IDContainer = id_container

        # Create the sample data for the patient
        sample_data = cx.SampleDataType()
        master_sample = cx.MasterSampleType()
        sample_id_container = cx.SampleIDContainerType()

        master_sample = cx.MasterSampleType(Source='XMLIMPORT',
            SampleTypeCatalogueTypeRef='TISSUE_OTH', SampleReceptacleTypeRef='ORG',
            hasChildren='false', SampleKind='TISSUE', SampleLocation='QBIC_STORAGE',
            SopDeviation='false')
        sample_id_container = cx.SampleIDContainerType()
        flex_id_1 = cx.FlexibleIDType(key='SAMPLEID', value=self._qbic_sample_id)
        flex_id_2 = cx.FlexibleIDType(key='QBIC_SAMPLE_ID', value=self._qbic_sample_id)

        # Add the flex IDs to the sample container
        sample_id_container.append(flex_id_1, flex_id_2)

        sampling_date = cx.DateType(Date=self._datetime, Precision='EXACT')

        init_amount = cx.VolumeType(Volume='1.0000', Unit='PC')

        master_sample.AmountRest = init_amount
        master_sample.InitialAmount = init_amount
        master_sample.SamplingDate = self._datetime
        master_sample.RepositionDate = self._datetime
        master_sample.FirstRepositionDate = self._datetime

        # Now extract the reference number of the dataset instances
        flex_dsref_list = []
        for flex_ds_instance in self._datasetinstance_list:
            ref = flex_ds_instance.reference()
            if not ref:
                raise NoReferenceIdFound("No reference id has been found"
                " for the given flexible dataset (SampleId: {})".format(self._qbic_sample_id))
        
        master_sample.FlexibleDataSetRef = flex_dsref_list
        patient_data_set.SampleData = master_sample

        return patient_data_set           



"""
A converter class, able to verify, extract and submit
important tumor diagnostic information from a ZIP archive
to CentraXX.
"""
import os
import sys
import uuid
from zipfile import ZipFile
from mtbparser.snv_utils import *
from .mtbfiletype import MtbFileType
from .mtbconverter_exceptions import MtbUnknownFileType
from .mtbconverter_exceptions import MtbIncompleteArchive
from .mtbconverter_exceptions import MtbUnknownBarcode
from .datasetinstance import DataSetInstance
from .patientdataset import PatientDataSet
from .cx_profiles import *
from mtbparser.snv_parser import SnvParser
from mtbparser.snv_utils import SSnvHeader
from datetime import datetime

TEMP_BASE_PATH = "/tmp"

FILETYPE_HEADER = {
    MtbFileType.GERM_SNV: GSnvHeader,
    MtbFileType.SOM_SNV: SSnvHeader,
    MtbFileType.GERM_CNV: CnvHeader,
    MtbFileType.SOM_CNV: CnvHeader,
    MtbFileType.SOM_SV: SsvHeader,
    MtbFileType.META: MetaData
    }

VERSIONS = {
    MtbFileType.GERM_SNV: 'QBIC_GERMLINE_SNV_V{}'.format(GERMLINE_SNV_VERSION) ,
    MtbFileType.SOM_SNV: 'QBIC_SOMATIC_SNV_V{}'.format(SOMATIC_SNV_VERSION),
    MtbFileType.GERM_CNV: 'QBIC_GERMLINE_CNV_V{}'.format(GERMLINE_CNV_VERSION),
    MtbFileType.SOM_CNV: 'QBIC_SOMATIC_CNV_V{}'.format(SOMATIC_CNV_VERSION),
    MtbFileType.SOM_SV: 'QBIC_SOMATIC_SV_V{}'.format(SOMATIC_SV_VERSION),
    MtbFileType.META: 'QBIC_METADATA_V{}'.format(METADATA_VERSION)
}

class MtbConverter():

    def __init__(self, zip_file, sample_code, patient_code):
        pyxb.utils.domutils.BindingDOMSupport.DeclareNamespace(cx.Namespace, 'cxx')
        self._root = cx.CentraXXDataExchange()
        self._zip_file = zip_file
        self._sample_code = sample_code
        self._patient_code = patient_code
        self._filelist = {}
        self._snvlists = {}
        self._tmp_dir = TEMP_BASE_PATH + os.path.sep + "tmp_uktdiagnostics_" + uuid.uuid4().hex
        os.mkdir(self._tmp_dir)
        self._verify()
        self._extract()
        self._cleanup()

    def _extract(self):
        try:
            self._zip_file.extractall(path=self._tmp_dir)
        except Exception as exc:
            print(exc)
            self._cleanup()
            sys.exit(1)
   
    def convert(self):
        print("Conversion started")
        flex_ds_instances = []
        datetime = datetime.isoformat(datetime.now())
        counter = 1

        # First, we have to build a flexible data type
        # for EVERY variant entry!
        for variant_file, mtb_filetype in self._filelist.items():
            file_path = self._tmp_dir + os.path.sep + variant_file
            print('Sample {}: Processing of file {}'.format(self._sample_code,
                variant_file))
            snv_list = SnvParser(file_path, FILETYPE_HEADER[mtb_filetype]).getSNVs()
            version = VERSIONS[mtb_filetype] 
            for snv_item in snv_list:
                ds_instance = DataSetInstance(
                    header_type=FILETYPE_HEADER[mtb_filetype],
                    snv_item=snv_item,
                    date=datetime,
                    version=version,
                    instance=counter
                )
                flex_ds_instances.append(ds_instance)
                counter += 1

        # Second, we have to build a patient data set
        # that consumes the references from the flexible
        # data set instances
        patient_ds = PatientDataSet(
            qbic_pat_id=self._patient_code,
            qbic_sample_id=self._sample_code,
            datetime=datetime,
            datasetinstance_list=flex_ds_instances
        )
        
        # Third, build the complete CentraXXDataExchange XML
        # object, that can be consumed by CentraXX
        flex_data_sets = [ds.datasetinstance() for ds in flex_ds_instances]
        self._root.Source = 'XMLIMPORT'
        effect_data = cx.EffectDataType(PatientDataSet=patient_ds.patientdataset())
        effect_data.append(flex_data_sets)
        self._root.EffectData = effect_data   
        
        root_dom = self._root.toDOM()
        root_dom.documentElement.setAttributeNS(
            xsi.uri(), 'xsi:schemaLocation',
            'http://www.kairos-med.de ../CentraXXExchange.xsd')
        root_dom.documentElement.setAttributeNS(
            xmlns.uri(), 'xmlns:xsi', xsi.uri())
        return root_dom.toprettyxml(encoding='utf-8')

    def _verify(self):
        for zipinfo in self._zip_file.infolist():
            filename = zipinfo.filename
            assigned_type = self._getfiletype(filename)
            if self._sample_code not in filename:
                raise MtbUnknownBarcode("Could not find sample_code {} in file '{}'. All"
                " files need to contain the same "
                "sample_code as the zip archive.".format(self._sample_code, filename))

            if not assigned_type:
                raise MtbUnknownFileType("Could not determine filetype "
                "according to the naming convention."
                " {} was probably not defined.".format(filename))
            self._filelist[filename] = assigned_type
        
        if len(self._filelist) != len(MtbFileType):
            raise MtbIncompleteArchive("The archive did not contain all necessary files."
            " Only found: {}".format(self._filelist.values()))
        
    def _cleanup(self):
        print(self._tmp_dir)
        os.removedirs(self._tmp_dir)

    def _getfiletype(self, filename):
        for filetype in MtbFileType:
            if filetype.value in filename:
                return filetype
        return ""



"""Builder for CentraXX measure profile XML"""

import pyxb.utils.domutils
import xml.dom.minidom
from pyxb.namespace import XMLSchema_instance as xsi
from pyxb.namespace import XMLNamespaces as xmlns
import mtbconverter.cxxpy as cx
from . import cx_utils as cxu

SOMATIC_SNV_VERSION = 1.0
SOMATIC_CNV_VERSION = 1.0
SOMATIC_SV_VERSION = 1.0
GERMLINE_SNV_VERSION = 1.0
GERMLINE_CNV_VERSION = 1.0
METADATA_VERSION = 1.0

SNV_FIELD_TYPES = ["QBIC_chr", "QBIC_start", "QBIC_ref", "QBIC_alt", "QBIC_allele_frequency_tumor",
"QBIC_coverage", "QBIC_gene", "QBIC_base_change", "QBIC_aa_change", "QBIC_transcript", "QBIC_functional_class",
"QBIC_effect"]

GSNV_FIELD_TYPES = ["QBIC_chr", "QBIC_start", "QBIC_ref", "QBIC_alt", "QBIC_genotype",
"QBIC_gene", "QBIC_base_change", "QBIC_aa_change", "QBIC_transcript", "QBIC_functional_class",
"QBIC_effect"]

SCNV_FIELD_TYPES = ["QBIC_size", "QBIC_type", "QBIC_copy_number", "QBIC_gene", "QBIC_exons", "QBIC_transcript",
"QBIC_chr", "QBIC_start", "QBIC_end", "QBIC_effect"]

GCNV_FIELD_TYPES = ["QBIC_size", "QBIC_type", "QBIC_copy_number", "QBIC_gene", "QBIC_exons", "QBIC_transcript",
"QBIC_chr", "QBIC_start", "QBIC_end", "QBIC_effect"]

SV_FIELD_TYPES = ["QBIC_type", "QBIC_gene", "QBIC_effect", "QBIC_left_bp", "QBIC_right_bp"]

METADATA_TYPES = ["QBIC_diagnosis", "QBIC_tumor_content", "QBIC_pathogenic_germ", "QBIC_mutational_load",
"QBIC_chrom_instability", "QBIC_quality_flag", "QBIC_reference_genome"]

class SSNVProfiles():

    def __init__(self):
        pyxb.utils.domutils.BindingDOMSupport.DeclareNamespace(cx.Namespace, 'cxx')
        self._root = cx.CentraXXDataExchange()
        self._source = 'XML-IMPORT'
        self._catalogue_data = cx.CatalogueDataType()
        self._profile()
    
    def _profile(self):
        """Build the profile for somatic snvs"""
        flexible_data_set = cx.FlexibleDataSetType(
            Code="QBIC_SOMATIC_SNV_V{:.1f}".format(SOMATIC_SNV_VERSION),
            Systemwide="false",
            FlexibleDataSetType="MEASUREMENT",
            Category="LABOR")
        multi_lang_de = cx.MultilingualEntryType(Lang='de', Value='Somatic SNV V{:.1f}'.format(SOMATIC_SNV_VERSION))
        multi_lang_en = cx.MultilingualEntryType(Lang='en', Value='Somatic SNV V{:.1f}'.format(SOMATIC_SNV_VERSION))
        flexible_data_set.NameMultilingualEntries = [multi_lang_de, multi_lang_en]
    
        flexible_data_set.FlexibleValueComplexRefs = [
            cx.FlexibleValueRefType(FlexibleValueRef=field_type, Required="false") for field_type in SNV_FIELD_TYPES
        ]
        
        self._catalogue_data.append(flexible_data_set)

        cfr_template = cx.CrfTemplateType(Name='Somatic SNV V{:.1f}'.format(SOMATIC_SNV_VERSION),
            FlexibleDataSetRef='QBIC_SOMATIC_SNV_V{:.1f}'.format(SOMATIC_SNV_VERSION),
            TemplateType='LABORMETHOD', Version='{:.0f}'.format(SOMATIC_SNV_VERSION-1), EntityStatus='ACTIVE', Global='false', MultipleUse='false', Active='false')

        cfr_template_section = cx.CrfTemplateSectionType(Name='Somatic SNV V{:.1f}'.format(SOMATIC_SNV_VERSION),
            Height=len(SNV_FIELD_TYPES), Width='1', Position='-1')
        
        cfr_template_section.CrfTemplateField = [
            cx.CrfTemplateFieldType(LaborValue=value,
                LowerRow=ind,
                LowerColumn='0',
                UpperRow=ind,
                UpperColumn='0',
                Mandatory='false',
                VisibleCaption='true',
                FieldType='LABORVALUE') for ind, value in enumerate(SNV_FIELD_TYPES)]

        cfr_template.CrfTemplateSection = [cfr_template_section]

        self._catalogue_data.append(cfr_template)       
        
    def getXML(self):
        """Write the XML object into an XML
        file"""
        self._root.Source = self._source
        self._root.CatalogueData = self._catalogue_data
        root_dom = self._root.toDOM()
        root_dom.documentElement.setAttributeNS(
            xsi.uri(), 'xsi:schemaLocation',
            'http://www.kairos-med.de ../CentraXXExchange.xsd')
        root_dom.documentElement.setAttributeNS(
            xmlns.uri(), 'xmlns:xsi', xsi.uri())
        return root_dom.toprettyxml(encoding='utf-8')


class GSNVProfiles():

    def __init__(self):
        pyxb.utils.domutils.BindingDOMSupport.DeclareNamespace(cx.Namespace, 'cxx')
        self._root = cx.CentraXXDataExchange()
        self._source = 'XML-IMPORT'
        self._catalogue_data = cx.CatalogueDataType()
        self._profile()
    
    def _profile(self):
        """Build the profile for germline snvs"""
        flexible_data_set = cx.FlexibleDataSetType(
            Code="QBIC_GERMLINE_SNV_V{:.1f}".format(GERMLINE_SNV_VERSION),
            Systemwide="false",
            FlexibleDataSetType="MEASUREMENT",
            Category="LABOR")
        multi_lang_de = cx.MultilingualEntryType(Lang='de', Value='Germline SNV V{:.1f}'.format(GERMLINE_SNV_VERSION))
        multi_lang_en = cx.MultilingualEntryType(Lang='en', Value='Germline SNV V{:.1f}'.format(GERMLINE_SNV_VERSION))
        flexible_data_set.NameMultilingualEntries = [multi_lang_de, multi_lang_en]      
    
        flexible_data_set.FlexibleValueComplexRefs = [
            cx.FlexibleValueRefType(FlexibleValueRef=field_type, Required="false") for field_type in GSNV_FIELD_TYPES
        ]
        
        self._catalogue_data.append(flexible_data_set)

        cfr_template = cx.CrfTemplateType(Name='Germline SNV V{:.1f}'.format(GERMLINE_SNV_VERSION),
            FlexibleDataSetRef='QBIC_GERMLINE_SNV_V{:.1f}'.format(GERMLINE_SNV_VERSION),
            TemplateType='LABORMETHOD', Version='{:.0f}'.format(GERMLINE_SNV_VERSION-1), EntityStatus='ACTIVE', Global='false', MultipleUse='false', Active='false')

        cfr_template_section = cx.CrfTemplateSectionType(Name='Germline SNV V{:.1f}'.format(GERMLINE_SNV_VERSION),
            Height=len(GSNV_FIELD_TYPES), Width='1', Position='-1')
        
        cfr_template_section.CrfTemplateField = [
            cx.CrfTemplateFieldType(LaborValue=value,
                LowerRow=ind,
                LowerColumn='0',
                UpperRow=ind,
                UpperColumn='0',
                Mandatory='false',
                VisibleCaption='true',
                FieldType='LABORVALUE') for ind, value in enumerate(GSNV_FIELD_TYPES)]

        cfr_template.CrfTemplateSection = [cfr_template_section]

        self._catalogue_data.append(cfr_template)       
        
    def getXML(self):
        """Write the XML object into an XML
        file"""
        self._root.Source = self._source
        self._root.CatalogueData = self._catalogue_data
        root_dom = self._root.toDOM()
        root_dom.documentElement.setAttributeNS(
            xsi.uri(), 'xsi:schemaLocation',
            'http://www.kairos-med.de ../CentraXXExchange.xsd')
        root_dom.documentElement.setAttributeNS(
            xmlns.uri(), 'xmlns:xsi', xsi.uri())
        return root_dom.toprettyxml(encoding='utf-8')

class SCNVProfiles():

    def __init__(self):
        pyxb.utils.domutils.BindingDOMSupport.DeclareNamespace(cx.Namespace, 'cxx')
        self._root = cx.CentraXXDataExchange()
        self._source = 'XML-IMPORT'
        self._catalogue_data = cx.CatalogueDataType()
        self._profile()
    
    def _profile(self):
        """Build the profile for somatic cnvs"""
        flexible_data_set = cx.FlexibleDataSetType(
            Code="QBIC_SOMATIC_CNV_V{:.1f}".format(SOMATIC_CNV_VERSION),
            Systemwide="false",
            FlexibleDataSetType="MEASUREMENT",
            Category="LABOR")
        multi_lang_de = cx.MultilingualEntryType(Lang='de', Value='Somatic CNV V{:.1f}'.format(SOMATIC_CNV_VERSION))
        multi_lang_en = cx.MultilingualEntryType(Lang='en', Value='Somatic CNV V{:.1f}'.format(SOMATIC_CNV_VERSION))
        flexible_data_set.NameMultilingualEntries = [multi_lang_de, multi_lang_en]      
    
        flexible_data_set.FlexibleValueComplexRefs = [
            cx.FlexibleValueRefType(FlexibleValueRef=field_type, Required="false") for field_type in SCNV_FIELD_TYPES
        ]
        
        self._catalogue_data.append(flexible_data_set)

        cfr_template = cx.CrfTemplateType(Name='Somatic CNV V{:.1f}'.format(SOMATIC_CNV_VERSION),
            FlexibleDataSetRef='QBIC_SOMATIC_CNV_V{:.1f}'.format(SOMATIC_CNV_VERSION),
            TemplateType='LABORMETHOD', Version='{:.0f}'.format(SOMATIC_CNV_VERSION-1), EntityStatus='ACTIVE', Global='false', MultipleUse='false', Active='false')

        cfr_template_section = cx.CrfTemplateSectionType(Name='Somatic CNV V{:.1f}'.format(SOMATIC_CNV_VERSION),
            Height=len(GSNV_FIELD_TYPES), Width='1', Position='-1')
        
        cfr_template_section.CrfTemplateField = [
            cx.CrfTemplateFieldType(LaborValue=value,
                LowerRow=ind,
                LowerColumn='0',
                UpperRow=ind,
                UpperColumn='0',
                Mandatory='false',
                VisibleCaption='true',
                FieldType='LABORVALUE') for ind, value in enumerate(SCNV_FIELD_TYPES)]

        cfr_template.CrfTemplateSection = [cfr_template_section]

        self._catalogue_data.append(cfr_template)       
        
    def getXML(self):
        """Write the XML object into an XML
        file"""
        self._root.Source = self._source
        self._root.CatalogueData = self._catalogue_data
        root_dom = self._root.toDOM()
        root_dom.documentElement.setAttributeNS(
            xsi.uri(), 'xsi:schemaLocation',
            'http://www.kairos-med.de ../CentraXXExchange.xsd')
        root_dom.documentElement.setAttributeNS(
            xmlns.uri(), 'xmlns:xsi', xsi.uri())
        return root_dom.toprettyxml(encoding='utf-8')

class GCNVProfiles():

    def __init__(self):
        pyxb.utils.domutils.BindingDOMSupport.DeclareNamespace(cx.Namespace, 'cxx')
        self._root = cx.CentraXXDataExchange()
        self._source = 'XML-IMPORT'
        self._catalogue_data = cx.CatalogueDataType()
        self._profile()
    
    def _profile(self):
        """Build the profile for germline cnvs"""
        flexible_data_set = cx.FlexibleDataSetType(
            Code="QBIC_GERMLINE_CNV_V{:.1f}".format(SOMATIC_CNV_VERSION),
            Systemwide="false",
            FlexibleDataSetType="MEASUREMENT",
            Category="LABOR")
        multi_lang_de = cx.MultilingualEntryType(Lang='de', Value='Germline CNV V{:.1f}'.format(GERMLINE_CNV_VERSION))
        multi_lang_en = cx.MultilingualEntryType(Lang='en', Value='Germline CNV V{:.1f}'.format(GERMLINE_CNV_VERSION))
        flexible_data_set.NameMultilingualEntries = [multi_lang_de, multi_lang_en]      
    
        flexible_data_set.FlexibleValueComplexRefs = [
            cx.FlexibleValueRefType(FlexibleValueRef=field_type, Required="false") for field_type in GCNV_FIELD_TYPES
        ]
        
        self._catalogue_data.append(flexible_data_set)

        cfr_template = cx.CrfTemplateType(Name='Germline CNV V{:.1f}'.format(GERMLINE_CNV_VERSION),
            FlexibleDataSetRef='QBIC_GERMLINE_CNV_V{:.1f}'.format(GERMLINE_CNV_VERSION),
            TemplateType='LABORMETHOD', Version='{:.0f}'.format(GERMLINE_CNV_VERSION-1), EntityStatus='ACTIVE', Global='false', MultipleUse='false', Active='false')

        cfr_template_section = cx.CrfTemplateSectionType(Name='Germline CNV V{:.1f}'.format(GERMLINE_CNV_VERSION),
            Height=len(GCNV_FIELD_TYPES), Width='1', Position='-1')
        
        cfr_template_section.CrfTemplateField = [
            cx.CrfTemplateFieldType(LaborValue=value,
                LowerRow=ind,
                LowerColumn='0',
                UpperRow=ind,
                UpperColumn='0',
                Mandatory='false',
                VisibleCaption='true',
                FieldType='LABORVALUE') for ind, value in enumerate(GCNV_FIELD_TYPES)]

        cfr_template.CrfTemplateSection = [cfr_template_section]

        self._catalogue_data.append(cfr_template)       
        
    def getXML(self):
        """Write the XML object into an XML
        file"""
        self._root.Source = self._source
        self._root.CatalogueData = self._catalogue_data
        root_dom = self._root.toDOM()
        root_dom.documentElement.setAttributeNS(
            xsi.uri(), 'xsi:schemaLocation',
            'http://www.kairos-med.de ../CentraXXExchange.xsd')
        root_dom.documentElement.setAttributeNS(
            xmlns.uri(), 'xmlns:xsi', xsi.uri())
        return root_dom.toprettyxml(encoding='utf-8')

class SVProfiles():

    def __init__(self):
        pyxb.utils.domutils.BindingDOMSupport.DeclareNamespace(cx.Namespace, 'cxx')
        self._root = cx.CentraXXDataExchange()
        self._source = 'XML-IMPORT'
        self._catalogue_data = cx.CatalogueDataType()
        self._profile()
    
    def _profile(self):
        """Build the profile for somatic structural variants"""
        flexible_data_set = cx.FlexibleDataSetType(
            Code="QBIC_SOMATIC_SV_V{:.1f}".format(SOMATIC_SV_VERSION),
            Systemwide="false",
            FlexibleDataSetType="MEASUREMENT",
            Category="LABOR")
        multi_lang_de = cx.MultilingualEntryType(Lang='de', Value='Somatic SV V{:.1f}'.format(SOMATIC_SV_VERSION))
        multi_lang_en = cx.MultilingualEntryType(Lang='en', Value='Somatic SV V{:.1f}'.format(SOMATIC_SV_VERSION))
        flexible_data_set.NameMultilingualEntries = [multi_lang_de, multi_lang_en]      
    
        flexible_data_set.FlexibleValueComplexRefs = [
            cx.FlexibleValueRefType(FlexibleValueRef=field_type, Required="false") for field_type in SV_FIELD_TYPES
        ]
        
        self._catalogue_data.append(flexible_data_set)

        cfr_template = cx.CrfTemplateType(Name='Somatic SV V{:.1f}'.format(SOMATIC_SV_VERSION),
            FlexibleDataSetRef='QBIC_SOMATIC_SV_V{:.1f}'.format(SOMATIC_SV_VERSION),
            TemplateType='LABORMETHOD', Version='{:.0f}'.format(SOMATIC_SV_VERSION-1), EntityStatus='ACTIVE', Global='false', MultipleUse='false', Active='false')

        cfr_template_section = cx.CrfTemplateSectionType(Name='Somatic SV V{:.1f}'.format(SOMATIC_SV_VERSION),
            Height=len(SV_FIELD_TYPES), Width='1', Position='-1')
        
        cfr_template_section.CrfTemplateField = [
            cx.CrfTemplateFieldType(LaborValue=value,
                LowerRow=ind,
                LowerColumn='0',
                UpperRow=ind,
                UpperColumn='0',
                Mandatory='false',
                VisibleCaption='true',
                FieldType='LABORVALUE') for ind, value in enumerate(SV_FIELD_TYPES)]

        cfr_template.CrfTemplateSection = [cfr_template_section]

        self._catalogue_data.append(cfr_template)       
        
    def getXML(self):
        """Write the XML object into an XML
        file"""
        self._root.Source = self._source
        self._root.CatalogueData = self._catalogue_data
        root_dom = self._root.toDOM()
        root_dom.documentElement.setAttributeNS(
            xsi.uri(), 'xsi:schemaLocation',
            'http://www.kairos-med.de ../CentraXXExchange.xsd')
        root_dom.documentElement.setAttributeNS(
            xmlns.uri(), 'xmlns:xsi', xsi.uri())
        return root_dom.toprettyxml(encoding='utf-8')

class MetadataProfiles():

    def __init__(self):
        pyxb.utils.domutils.BindingDOMSupport.DeclareNamespace(cx.Namespace, 'cxx')
        self._root = cx.CentraXXDataExchange()
        self._source = 'XML-IMPORT'
        self._catalogue_data = cx.CatalogueDataType()
        self._profile()
    
    def _profile(self):
        """Build the profile for metadata information"""
        flexible_data_set = cx.FlexibleDataSetType(
            Code="QBIC_METADATA_V{:.1f}".format(METADATA_VERSION),
            Systemwide="false",
            FlexibleDataSetType="MEASUREMENT",
            Category="LABOR")
        multi_lang_de = cx.MultilingualEntryType(Lang='de', Value='Metadata V{:.1f}'.format(METADATA_VERSION))
        multi_lang_en = cx.MultilingualEntryType(Lang='en', Value='Metadata V{:.1f}'.format(METADATA_VERSION))
        flexible_data_set.NameMultilingualEntries = [multi_lang_de, multi_lang_en]      
    
        flexible_data_set.FlexibleValueComplexRefs = [
            cx.FlexibleValueRefType(FlexibleValueRef=field_type, Required="false") for field_type in METADATA_TYPES
        ]
        
        self._catalogue_data.append(flexible_data_set)

        cfr_template = cx.CrfTemplateType(Name='Metadata V{:.1f}'.format(METADATA_VERSION),
            FlexibleDataSetRef='QBIC_METADATA_V{:.1f}'.format(METADATA_VERSION),
            TemplateType='LABORMETHOD', Version='{:.0f}'.format(METADATA_VERSION-1), EntityStatus='ACTIVE', Global='false', MultipleUse='false', Active='false')

        cfr_template_section = cx.CrfTemplateSectionType(Name='Metadata V{:.1f}'.format(METADATA_VERSION),
            Height=len(METADATA_TYPES), Width='1', Position='-1')
        
        cfr_template_section.CrfTemplateField = [
            cx.CrfTemplateFieldType(LaborValue=value,
                LowerRow=ind,
                LowerColumn='0',
                UpperRow=ind,
                UpperColumn='0',
                Mandatory='false',
                VisibleCaption='true',
                FieldType='LABORVALUE') for ind, value in enumerate(METADATA_TYPES)]

        cfr_template.CrfTemplateSection = [cfr_template_section]

        self._catalogue_data.append(cfr_template)       
        
    def getXML(self):
        """Write the XML object into an XML
        file"""
        self._root.Source = self._source
        self._root.CatalogueData = self._catalogue_data
        root_dom = self._root.toDOM()
        root_dom.documentElement.setAttributeNS(
            xsi.uri(), 'xsi:schemaLocation',
            'http://www.kairos-med.de ../CentraXXExchange.xsd')
        root_dom.documentElement.setAttributeNS(
            xmlns.uri(), 'xmlns:xsi', xsi.uri())
        return root_dom.toprettyxml(encoding='utf-8')





"""
An enum class for having a controlled
vocabulary for the different file types.
"""
from enum import Enum

class MtbFileType(Enum):
    """Maps an unique mtb file type to an enum object"""
    GERM_SNV = "_germline_snv.tsv"
    GERM_CNV = "_germline_cnv.tsv"
    SOM_SNV = "_somatic_snv.tsv"
    SOM_CNV = "_somatic_cnv.tsv"
    SOM_SV = "_somatic_sv.tsv"
    META = "_metadata.tsv"



"""Builder for controlled vocabulary XML as part of the preparation
for the CentraXX import"""

import pyxb.utils.domutils
import xml.dom.minidom
from pyxb.namespace import XMLSchema_instance as xsi
from pyxb.namespace import XMLNamespaces as xmlns
import mtbconverter.cxxpy as cx
from mtbparser.snv_utils import SSnvHeader
from . import cx_utils as cxu

CV_VERSION = '0.1'

class ControlledVocabulary:
    
    def __init__(self):
        pyxb.utils.domutils.BindingDOMSupport.DeclareNamespace(cx.Namespace, 'cxx')
        self._root = cx.CentraXXDataExchange()
        self._source = 'XML-IMPORT'
        self._catalogue_data = cx.CatalogueDataType()
        self._cvforssnv()
    
    def _cvforssnv(self):
        """Create the XML entries for the
        somatic SNVs information"""

        # register the chromosomes (de, en)
        for chr_index in cxu.CV_CHROMOSOMES:
            multi_entry_type_de = cx.MultilingualEntryType(Lang='de', Value='Chromosome {}{}{}'.format(cxu.CV_PREFIX, 'chr',chr_index))
            multi_entry_type_en = cx.MultilingualEntryType(Lang='en', Value='Chromosome {}{}{}'.format(cxu.CV_PREFIX, 'chr',chr_index))
            item = cx.UsageEntryType()
            item.NameMultilingualEntries = [multi_entry_type_de, multi_entry_type_en]
            item.Code = '{}{}{}'.format(cxu.CV_PREFIX, 'chr',chr_index)
            item.Category = "false"
            self._catalogue_data.append(item)
        
        # create vc for sSNV effects
        for effect in cxu.CV_EFFECT_SSNV:
            multi_entry_type_de = cx.MultilingualEntryType(Lang='de', Value='{}{}'.format(cxu.CV_PREFIX, effect))
            multi_entry_type_en = cx.MultilingualEntryType(Lang='en', Value='{}{}'.format(cxu.CV_PREFIX, effect))
            item = cx.UsageEntryType()
            item.NameMultilingualEntries = [multi_entry_type_de, multi_entry_type_en]
            item.Code = '{}{}'.format(cxu.CV_PREFIX, effect)
            item.Category = "false"
            self._catalogue_data.append(item)
        
        # create cv for bases
        for base in cxu.CV_BASES:
            multi_entry_type_de = cx.MultilingualEntryType(Lang='de', Value='{}{}{}'.format(cxu.CV_PREFIX, 'base', base))
            multi_entry_type_en = cx.MultilingualEntryType(Lang='en', Value='{}{}{}'.format(cxu.CV_PREFIX, 'base', base))
            item = cx.UsageEntryType()
            item.NameMultilingualEntries = [multi_entry_type_de, multi_entry_type_en]
            item.Code = '{}{}{}'.format(cxu.CV_PREFIX, 'base', base)
            item.Category = "false"
            self._catalogue_data.append(item)
        
        # create cv for mutational load
        for load in cxu.CV_MUT_LOAD:
            multi_entry_type_de = cx.MultilingualEntryType(Lang='de', Value='{}{}_{}'.format(cxu.CV_PREFIX, 'mutational_load', load))
            multi_entry_type_en = cx.MultilingualEntryType(Lang='en', Value='{}{}_{}'.format(cxu.CV_PREFIX, 'mutational_load', load))
            item = cx.UsageEntryType()
            item.NameMultilingualEntries = [multi_entry_type_de, multi_entry_type_en]
            item.Code = '{}{}_{}'.format(cxu.CV_PREFIX, 'mutational_load', load)
            item.Category = "false"
            self._catalogue_data.append(item)
        
        for genotype in cxu.CV_GENOTYPES:
            multi_entry_type_de = cx.MultilingualEntryType(Lang='de', Value='{}{}_{}'.format(cxu.CV_PREFIX, 'genotype', genotype))
            multi_entry_type_en = cx.MultilingualEntryType(Lang='en', Value='{}{}_{}'.format(cxu.CV_PREFIX, 'genotype', genotype))
            item = cx.UsageEntryType()
            item.NameMultilingualEntries = [multi_entry_type_de, multi_entry_type_en]
            item.Code = '{}{}_{}'.format(cxu.CV_PREFIX, 'genotype', genotype)
            item.Category = "false"
            self._catalogue_data.append(item)

        
    def getXML(self):
        """Write the XML object into an XML
        file"""
        self._root.Source = self._source
        self._root.CatalogueData = self._catalogue_data
        root_dom = self._root.toDOM()
        root_dom.documentElement.setAttributeNS(
            xsi.uri(), 'xsi:schemaLocation',
            'http://www.kairos-med.de ../CentraXXExchange.xsd')
        root_dom.documentElement.setAttributeNS(
            xmlns.uri(), 'xmlns:xsi', xsi.uri())
        return root_dom.toprettyxml(encoding='utf-8')




"""
Utility collection with helper methods
"""
import os
import re

BARCODE_REGEX = "Q[A-X0-9]{4}[0-9]{3}[A-X][A-X0-9]"

def getbarcode(path):
    """Extract an OpenBis barcode from the file name.
    If a barcode is found, return it. Raise ValueError if no barcode,
    or more that one barcode has been found.
    Barcodes must match this regular expression: [A-Z]{5}[0-9]{3}[A-Z][A-Z0-9]
    """
    stem, suffix = os.path.splitext(os.path.basename(path))
    barcodes = re.findall(BARCODE_REGEX, stem)
    valid_barcodes = [b for b in barcodes if is_valid_barcode(b)]
    if not barcodes:
        raise ValueError("no barcodes found")
    if len(set(barcodes)) > 1:
        raise ValueError("more than one barcode in filename")
    return barcodes[0]


def is_valid_barcode(barcode):
    """Check if barcode is a valid OpenBis barcode."""
    if re.match('^' + BARCODE_REGEX + '$', barcode) is None:
        return False
    csum = sum(ord(c) * (i + 1) for i, c in enumerate(barcode[:-1]))
    csum = csum % 34 + 48
    if csum > 57:
        csum += 7
    if barcode[-1] == chr(csum):
        return True
    return False